@@ -39,9 +39,12 @@
 
 #include "sccp_internal.h"
 #include "xua_internal.h"
+#include "ss7_as.h"
+#include "ss7_asp.h"
 #include "ss7_route.h"
 #include "ss7_route_table.h"
 #include "ss7_internal.h"
+#include "ss7_xua_srv.h"
 
 /*! \brief Find a SCCP User registered for given PC+SSN or SSN only
  * First search all users with a valid PC for a full PC+SSN match.
@@ -89,9 +92,14 @@ osmo_sccp_user_find(struct osmo_sccp_instance *inst, uint16_t ssn, uint32_t pc)
 /*! \brief Bind a SCCP User to a given Point Code
  *  \param[in] inst SCCP Instance
  *  \param[in] name human-readable name
+ *  \param[in] prim_cb User provided callback to pass a primitive/msg up the stack
  *  \param[in] ssn Sub-System Number to bind to
  *  \param[in] pc Point Code to bind to, or OSMO_SS7_PC_INVALID if none.
- *  \returns Callee-allocated SCCP User on success; negative otherwise */
+ *  \returns Callee-allocated SCCP User on success; negative otherwise
+ *
+ * Ownership of oph->msg in prim_cb is transferred to the user of the
+ * registered callback when called.
+ */
 static struct osmo_sccp_user *
 sccp_user_bind_pc(struct osmo_sccp_instance *inst, const char *name,
 		  osmo_prim_cb prim_cb, uint16_t ssn, uint32_t pc)
@@ -124,9 +132,14 @@ sccp_user_bind_pc(struct osmo_sccp_instance *inst, const char *name,
 /*! \brief Bind a given SCCP User to a given SSN+PC
  *  \param[in] inst SCCP Instance
  *  \param[in] name human-readable name
+ *  \param[in] prim_cb User provided callback to pass a primitive/msg up the stack
  *  \param[in] ssn Sub-System Number to bind to
  *  \param[in] pc Point Code to bind to
- *  \returns Callee-allocated SCCP User on success; negative otherwise */
+ *  \returns Callee-allocated SCCP User on success; negative otherwise
+ *
+ * Ownership of oph->msg in prim_cb is transferred to the user of the
+ * registered callback when called.
+ */
 struct osmo_sccp_user *
 osmo_sccp_user_bind_pc(struct osmo_sccp_instance *inst, const char *name,
 		       osmo_prim_cb prim_cb, uint16_t ssn, uint32_t pc)
@@ -137,8 +150,13 @@ osmo_sccp_user_bind_pc(struct osmo_sccp_instance *inst, const char *name,
 /*! \brief Bind a given SCCP User to a given SSN (at any PC)
  *  \param[in] inst SCCP Instance
  *  \param[in] name human-readable name
+ *  \param[in] prim_cb User provided callback to pass a primitive/msg up the stack
  *  \param[in] ssn Sub-System Number to bind to
- *  \returns Callee-allocated SCCP User on success; negative otherwise */
+ *  \returns Callee-allocated SCCP User on success; negative otherwise
+ *
+ * Ownership of oph->msg in prim_cb is transferred to the user of the
+ * registered callback when called.
+ */
 struct osmo_sccp_user *
 osmo_sccp_user_bind(struct osmo_sccp_instance *inst, const char *name,
 		    osmo_prim_cb prim_cb, uint16_t ssn)
@@ -172,7 +190,10 @@ void *osmo_sccp_user_get_priv(struct osmo_sccp_user *scu)
 /*! \brief Send a SCCP User SAP Primitive up to the User
  *  \param[in] scu SCCP User to whom to send the primitive
  *  \param[in] prim Primitive to send to the user
- *  \returns return value of the SCCP User's prim_cb() function */
+ *  \returns return value of the SCCP User's prim_cb() function
+ *
+ * Ownership of prim->oph->msg is passed to the user of the registered callback
+ */
 int sccp_user_prim_up(struct osmo_sccp_user *scu, struct osmo_scu_prim *prim)
 {
 	LOGP(DLSCCP, LOGL_DEBUG, "Delivering %s to SCCP User '%s'\n",
@@ -234,10 +255,6 @@ osmo_sccp_instance_create(struct osmo_ss7_instance *ss7, void *priv)
 	inst->priv = priv;
 	INIT_LLIST_HEAD(&inst->users);
 
-	inst->ss7_user.inst = ss7;
-	inst->ss7_user.name = "SCCP";
-	inst->ss7_user.prim_cb = mtp_user_prim_cb;
-	inst->ss7_user.priv = inst;
 	inst->max_optional_data = SCCP_MAX_OPTIONAL_DATA;
 
 	inst->tdefs = talloc_memdup(inst, osmo_sccp_timer_defaults,
@@ -250,7 +267,10 @@ osmo_sccp_instance_create(struct osmo_ss7_instance *ss7, void *priv)
 		return NULL;
 	}
 
-	osmo_ss7_user_register(ss7, MTP_SI_SCCP, &inst->ss7_user);
+	inst->ss7_user = osmo_ss7_user_create(ss7, "SCCP");
+	osmo_ss7_user_set_prim_cb(inst->ss7_user, mtp_user_prim_cb);
+	osmo_ss7_user_set_priv(inst->ss7_user, inst);
+	osmo_ss7_user_register(ss7, MTP_SI_SCCP, inst->ss7_user);
 
 	llist_add_tail(&inst->list, &sccp_instances);
 
@@ -262,7 +282,9 @@ void osmo_sccp_instance_destroy(struct osmo_sccp_instance *inst)
 	struct osmo_sccp_user *scu, *scu2;
 
 	inst->ss7->sccp = NULL;
-	osmo_ss7_user_unregister(inst->ss7, MTP_SI_SCCP, &inst->ss7_user);
+	osmo_ss7_user_unregister(inst->ss7, MTP_SI_SCCP, inst->ss7_user);
+	osmo_ss7_user_destroy(inst->ss7_user);
+	inst->ss7_user = NULL;
 
 	llist_for_each_entry_safe(scu, scu2, &inst->users, list) {
 		osmo_sccp_user_unbind(scu);
@@ -487,6 +509,18 @@ const char *osmo_sccp_user_name(struct osmo_sccp_user *scu)
  * Convenience function for CLIENT
  ***********************************************************************/
 
+ /* Returns whether AS is already associated to any AS.
+  * Helper function for osmo_sccp_simple_client_on_ss7_id(). */
+static bool asp_serves_some_as(const struct osmo_ss7_asp *asp)
+{
+	struct osmo_ss7_as *as_i;
+	llist_for_each_entry(as_i, &asp->inst->as_list, list) {
+		if (osmo_ss7_as_has_asp(as_i, asp))
+			return true;
+	}
+	return false;
+}
+
 /*! \brief request an sccp client instance
  *  \param[in] ctx talloc context
  *  \param[in] ss7_id of the SS7/CS7 instance
@@ -606,24 +640,23 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 		/* Check if the user has created an ASP for this proto that is not added on any AS yet. */
 		struct osmo_ss7_asp *asp_i;
 		llist_for_each_entry(asp_i, &ss7->asp_list, list) {
-			struct osmo_ss7_as *as_i;
-			bool is_on_as = false;
 			if (asp_i->cfg.proto != prot)
 				continue;
-			llist_for_each_entry(as_i, &ss7->as_list, list) {
-				if (!osmo_ss7_as_has_asp(as_i, asp_i))
-					continue;
-				is_on_as = true;
-				break;
-			}
-			if (is_on_as) {
-				/* This ASP is already on another AS. If it was on this AS, we'd have found it above. */
+			if (asp_serves_some_as(asp_i)) {
+				/* This ASP is already on another AS.
+				 * If it was on this AS, we'd have found it above. */
 				continue;
 			}
 			/* This ASP matches the protocol and is not yet associated to any AS. Use it. */
 			asp = asp_i;
 			LOGP(DLSCCP, LOGL_NOTICE, "%s: ASP %s for %s is not associated with any AS, using it\n",
 			     name, asp->cfg.name, osmo_ss7_asp_protocol_name(prot));
+			ss7_as_add_asp(as, asp);
+			/* ASP became associated to a new AS, hence it needs to be
+			 * restarted to announce/register its Routing Context.
+			 * Make sure proper defaults are applied if app didn't
+			 * provide specific default values, then restart the ASP: */
+			ss7_asp_restart_after_reconfigure(asp);
 			break;
 		}
 		if (!asp) {
@@ -638,20 +671,20 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 			if (!asp)
 				goto out_rt;
 			asp_created = true;
+			asp->simple_client_allocated = true;
 			/* Ensure that the ASP we use is set to operate as a client. */
 			asp->cfg.is_server = false;
 			/* Ensure that the ASP we use is set to role ASP. */
 			asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
 			if (default_local_ip)
-				osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp, &default_local_ip, 1);
+				ss7_asp_peer_set_hosts(&asp->cfg.local, asp, &default_local_ip, 1);
 			if (default_remote_ip)
-				osmo_ss7_asp_peer_set_hosts(&asp->cfg.remote, asp, &default_remote_ip, 1);
-			/* Make sure proper defaults are applied if app didn't provide specific default values */
-			ss7_asp_set_default_peer_hosts(asp);
-			asp->simple_client_allocated = true;
+				ss7_asp_peer_set_hosts(&asp->cfg.remote, asp, &default_remote_ip, 1);
+			ss7_as_add_asp(as, asp);
+			/* Make sure proper defaults are applied if app didn't
+			provide specific default values, then restart the ASP: */
+			ss7_asp_restart_after_reconfigure(asp);
 		}
-
-		osmo_ss7_as_add_asp(as, asp->cfg.name);
 	}
 
 	/* Extra sanity checks if the ASP asp-clnt-* was pre-configured over VTY: */
@@ -682,7 +715,7 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 			LOGP(DLSCCP, LOGL_NOTICE,
 			     "%s: Requesting an SCCP simple client on ASP %s configured with 'transport-role server'\n",
 			     name, asp->cfg.name);
-			xs = osmo_ss7_xua_server_find2(ss7,
+			xs = ss7_xua_server_find2(ss7,
 						       asp->cfg.trans_proto, prot,
 						       asp->cfg.local.port);
 			if (!xs) {
@@ -692,12 +725,9 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 				goto out_asp;
 			}
 		}
+		/* ASP was already started here previously by VTY go_parent. */
 	}
 
-	/* Restart ASP */
-	if (prot != OSMO_SS7_ASP_PROT_IPA)
-		osmo_ss7_asp_use_default_lm(asp, LOGL_DEBUG);
-	osmo_ss7_asp_restart(asp);
 	LOGP(DLSCCP, LOGL_NOTICE, "%s: Using ASP instance %s\n", name,
 	     asp->cfg.name);
 
@@ -775,11 +805,11 @@ osmo_sccp_simple_server_on_ss7_id(void *ctx, uint32_t ss7_id, uint32_t pc,
 		return NULL;
 	ss7->cfg.primary_pc = pc;
 
-	xs = osmo_ss7_xua_server_create2(ss7, trans_proto, prot, local_port, local_ip);
+	xs = ss7_xua_server_create2(ss7, trans_proto, prot, local_port, local_ip);
 	if (!xs)
 		goto out_ss7;
 
-	rc = osmo_ss7_xua_server_bind(xs);
+	rc = ss7_xua_server_bind(xs);
 	if (rc < 0)
 		goto out_xs;
 
@@ -791,7 +821,7 @@ osmo_sccp_simple_server_on_ss7_id(void *ctx, uint32_t ss7_id, uint32_t pc,
 	return ss7->sccp;
 
 out_xs:
-	osmo_ss7_xua_server_destroy(xs);
+	ss7_xua_server_destroy(xs);
 out_ss7:
 	osmo_ss7_instance_destroy(ss7);
 
@@ -848,18 +878,18 @@ osmo_sccp_simple_server_add_clnt(struct osmo_sccp_instance *inst,
 					   trans_proto, prot);
 	if (!asp)
 		goto out_rt;
-	oxs = osmo_ss7_xua_server_find2(ss7, asp->cfg.trans_proto, prot, local_port);
+	oxs = ss7_xua_server_find2(ss7, asp->cfg.trans_proto, prot, local_port);
 	if (!oxs)
 		goto out_asp;
-	if (osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp,
+	if (ss7_asp_peer_set_hosts(&asp->cfg.local, asp,
 					(const char* const*)oxs->cfg.local.host,
 					oxs->cfg.local.host_cnt) < 0)
 		goto out_asp;
-	if (osmo_ss7_asp_peer_add_host(&asp->cfg.remote, asp, remote_ip) < 0)
+	if (ss7_asp_peer_add_host(&asp->cfg.remote, asp, remote_ip) < 0)
 		goto out_asp;
 	asp->cfg.is_server = true;
 	asp->cfg.role = OSMO_SS7_ASP_ROLE_SG;
-	osmo_ss7_as_add_asp(as, asp_name);
+	ss7_as_add_asp(as, asp);
 	talloc_free(asp_name);
 	talloc_free(as_name);
 	osmo_ss7_asp_restart(asp);
